/* Copyright (C) 2001  J.A. Bezemer  -  GPL >=2  -  NO WARRANTY WHATSOEVER */
/* The "trivial" parts are ripped from both libsafe and vsound. */

#include <stdio.h>		/* defines stdin */
#include <signal.h>

#include <dlfcn.h>		/* defines dlsym() */

#ifndef RTLD_NEXT
#define RTLD_NEXT  ((void *) -1l)
#endif
#define REAL_LIBC RTLD_NEXT


void *lowheap = (void *) 0x6fffffff, *highheap = NULL;



typedef void *(*memcpy_t) (void *dest, const void *src, size_t n);

void *
memcpy (void *dest, const void *src, size_t n)
{
  static memcpy_t real_memcpy = NULL;

  const unsigned char *s = src;

  if (!real_memcpy)
    real_memcpy = (memcpy_t) dlsym (REAL_LIBC, "memcpy");

  /* Fix for SEGV (don't ask) */
  if (n == 0x10000)
    n = 53300;

  /* Remember end addresses */
  if (dest > (void *) 0x7000000 && dest < (void *) 0xf000000
      && dest < lowheap)
    lowheap = dest;
  if (src > (void *) 0x7000000 && src < (void *) 0xf000000 && src < lowheap)
    lowheap = src;
  if (dest + n > (void *) 0x7000000 && dest + n < (void *) 0xf000000
      && dest + n > highheap)
    highheap = dest + n;
  if (src + n > (void *) 0x7000000 && src + n < (void *) 0xf000000
      && src + n > highheap)
    highheap = src + n;

  if (s[0] == ' ' && s[1] == '0' && s[2] == ' ')
    {
      fwrite (lowheap, highheap - lowheap, 1, stderr);
    }

  return real_memcpy (dest, src, n);
}



typedef void *(*memmove_t) (void *dest, const void *src, size_t n);

void *
memmove (void *dest, const void *src, size_t n)
{
  static memmove_t real_memmove = NULL;

  if (!real_memmove)
    real_memmove = (memmove_t) dlsym (REAL_LIBC, "memmove");

  /* Remember end addresses */
  if (dest > (void *) 0x7000000 && dest < (void *) 0xf000000
      && dest < lowheap)
    lowheap = dest;
  if (src > (void *) 0x7000000 && src < (void *) 0xf000000 && src < lowheap)
    lowheap = src;
  if (dest + n > (void *) 0x7000000 && dest + n < (void *) 0xf000000
      && dest + n > highheap)
    highheap = dest + n;
  if (src + n > (void *) 0x7000000 && src + n < (void *) 0xf000000
      && src + n > highheap)
    highheap = src + n;

  return real_memmove (dest, src, n);
}



typedef void *(*malloc_t) (size_t n);

void *
malloc (size_t n)
{
  static malloc_t real_malloc = NULL;
  void *res;

  if (!real_malloc)
    real_malloc = (malloc_t) dlsym (REAL_LIBC, "malloc");

  res = real_malloc (n);

  /* Remember end addresses */
  if (res > (void *) 0x7000000 && res < (void *) 0xf000000 && res < lowheap)
    lowheap = res;
  if (res + n > (void *) 0x7000000 && res + n < (void *) 0xf000000
      && res + n > highheap)
    highheap = res + n;

  return res;
}
